@@ -210,7 +210,8 @@ static struct crush_map *crush_decode(void *pbyval, void *end)
 		ceph_decode_need(p, end, 4*sizeof(u32), bad);
 		b->id = ceph_decode_32(p);
 		b->type = ceph_decode_16(p);
-		b->alg = ceph_decode_16(p);
+		b->alg = ceph_decode_8(p);
+		b->hash = ceph_decode_8(p);
 		b->weight = ceph_decode_32(p);
 		b->size = ceph_decode_32(p);
 
@@ -648,7 +649,7 @@ struct ceph_osdmap *osdmap_apply_incremental(void **p, void *end,
 		u32 osd;
 		ceph_decode_32_safe(p, end, osd, bad);
 		(*p)++;  /* clean flag */
-		pr_info("ceph osd%d down\n", osd);
+		pr_info("osd%d down\n", osd);
 		if (osd < map->max_osd)
 			map->osd_state[osd] &= ~CEPH_OSD_UP;
 	}
@@ -809,7 +810,7 @@ int ceph_calc_object_layout(struct ceph_object_layout *ol,
 		return -EIO;
 
 	pool = &osdmap->pg_pool[poolid];
-	ps = ceph_full_name_hash(oid, strlen(oid));
+	ps = ceph_str_hash(pool->v.object_hash, oid, strlen(oid));
 	if (preferred >= 0) {
 		ps += preferred;
 		num = le32_to_cpu(pool->v.lpg_num);
